function [struct_data, X] = mules_csv_parser_new(csv_filename, timestamps_bool)
%MULES_PARSER Reads CSV files created by MuLES and retrieves: 
% DATA, a structure with the device information and signals data
% CSV_DATA, a matrix where the 1st column is timestamps (if available) and 
%           the last column are the markers
%
% CSV_FILENAME  = CSV file name
% TIMESTAMPS    = Boolean to indicate if the CSV file has UTC timestamps.
%                 Default = True
%
% Raymundo Cassani
% 2016

% Verify the TIMESTAMPS argument
if ~exist('timestamps_bool','var') || isempty(timestamps_bool)
    timestamps_bool = true;
end


%% Read data from CSV file
fid = fopen(csv_filename);
% read Header
header_string =  fgetl(fid);
% get information of the hardware used
hardware = findfield_inheader(header_string, 'HARDWARE');
% dead Channel Names
channel_string = fgetl(fid);
tmp = textscan(channel_string,'%s','delimiter',',');
channel_labels = tmp{1};

% initialize some variables
i_row = 1;
marker_vct = [];
timestamps = [];

% Reads CSV file row by row
% (as row can have different size)
while true
    tmp = fgetl(fid);
    if ischar(tmp)
        % Comma-Separated-Values format to numeric array
        tmp2 = textscan(tmp, '%f', 'delimiter',',');
        data_row = tmp2{1,1}'; 
        % extract last element for each row, this is the Marker
        marker_vct(i_row) = data_row(end);
        data_row(end) = [];
        % extract the first element for each row, if it's MuLES timestamp
        if timestamps_bool
           timestamps(i_row) = data_row(1);
           data_row(1) = []; 
        end
        % store all the data, divided per row
        X_cell{i_row} = data_row;      
        i_row = i_row + 1;
    else
        %n_rows = i_row - 1;
        fclose(fid);
        break;
   end
end

% Data from CSV file
csv_data.header = header_string;
csv_data.channel_names = channel_labels;
csv_data.timestamps = timestamps' / 1000;
csv_data.X = X_cell';
csv_data.marker_vct = marker_vct';



% Initialize array X to contain Data, and marker_vct for Markers
% X = zeros(n_rows, n_cols);
%marker_vct = zeros(n_rows,1);
% % % open CSV file
% % fid = fopen(csv_filename);
% % % Headers
% % fgetl(fid);
% % % Channel names
% % fgetl(fid);

% % % for i_row = 1: n_rows
% % %    tmp = fgetl(fid);
% % %    % Comma separated values to numeric array
% % %    tmp2 = textscan(tmp, '%f', 'delimiter',',');
% % %    tmp3 = tmp2{1,1}'; 
% % %    marker_vct(i_row) = tmp3(end);
% % %    tmp_row = tmp3(1:end-1);
% % %    n_elem = numel(tmp_row);
% % %    X(i_row, 1:n_elem) = tmp_row; 
% % % end
% % % fclose(fid);
%%

% Full Data matrix
%X = [X, marker_vct];

%return;

% Parse and order data according to the Hardware
switch hardware
    case 'INTERAXON-MUSE'      
        struct_data = parse_csv_interaxon_muse(csv_data);
        
    case 'NEUROELECTRICS-ENOBIO'
        struct_data = parse_csv_neuroelectrics_enobio(csv_data);
         
    case 'EMPATICA-WINSERVER'
        struct_data = parse_csv_empatica_winserver(csv_data); 
        
    case 'ZEPHYR-BH3RAW'       
        struct_data = parse_csv_zephyr_bh3raw(csv_data);     
    
    case 'TOBII-EYEX'
        struct_data = parse_csv_tobii_eyex(csv_data); 
    
    case 'INTERAXON-MUSERAW'
        struct_data = parse_csv_interaxon_museraw(csv_data);
    
    otherwise
        struct_data = [];
        return;
end 

end

       
    function str_out = findfield_inheader(header_string, field)
    % Scans the header_string and gets the value for a given field
    % value is returned as string
        tmp = textscan(header_string,'%s','delimiter',',');
        cell_header = tmp{1};
        k = strncmp(cell_header, field, numel(field));
        tmpcell = (textscan(char(cell_header(k)),[field '=%s']));
        str_out = char(tmpcell{1});
    end
        
    function marker_vct = project_markers(main_time_vct, main_marker_vct, time_vct)
    % Given an main_marker_vector with its main_time_vector (same size), the markers
    % are projected into a new time vector
        marker_vct = time_vct*0;
        % Finds the indexes for the markers different from Zero 
        [marker_ix, ~, marker_val] = find(main_marker_vct);
        n_markers = numel(marker_ix);
        % Given of the original marker, finds it location in time_vct
        for i_marker = 1 : n_markers
            diff_time = abs(time_vct - main_time_vct(marker_ix(i_marker)));
            [~, ix] = min(diff_time);
            marker_vct(ix) = marker_val(i_marker);
        end
    end
    
    function column_vector = get_columns_from_cell(cell_array, ix_column )
    % Given a cell_array with n_r elements (n_r, 1), each element, r, has
    % the shape (1, n) where n can be different for each 'r' element in 
    % the cell array
    % get_columns_from_cell(cell_array, ix_column) extracts the ix_column
    % from each 'r' element from the cell array and concatenate those 
    % elements
    column_vector = cell2mat(cellfun(@(x) x(ix_column), cell_array, 'UniformOutput', false)); 
    end

    function data = parse_csv_interaxon_muse(csv_data)
    % Parses the CSV_DATA structure with
    % the HARDWARE = MUSE-INTERAXON
        % Transform X_cell to matrix X, sa each row has the same number of elements
        X = cell2mat(csv_data.X);
        % Get device name
        data.device_name = findfield_inheader(csv_data.header, 'NAME');
        % Get information of the hardware used
        data.hardware = findfield_inheader(csv_data.header, 'HARDWARE');
        % Channel names
        channel_labels = csv_data.channel_names;
        % timestamps and duration
        if ~isempty(csv_data.timestamps)
            timestamps = csv_data.timestamps;
            data.duration = timestamps(end) - timestamps(1);
            channel_labels(1) = [];
        end
        marker_vct = csv_data.marker_vct;
        channel_labels(end) = [];

        % Get MUSE Preset
        muse_preset = findfield_inheader(csv_data.header, 'PRESET');
        switch muse_preset
            case {'ab'}
                data.eeg.fs = 500;
            case {'13','14'}
                data.eeg.fs = 220;
        end        
        % If present, save timestamps
        if ~isempty(csv_data.timestamps)
            data.eeg.timestamps = timestamps;
            data.eeg.expected_samples = round(data.duration * data.eeg.fs);
        end
        data.eeg.acquired_samples = size(X, 1);
        data.eeg.time_vct = (0:data.eeg.acquired_samples-1)'/data.eeg.fs;
        data.eeg.markers = marker_vct;

        % Is Acceleration data present?
        acc_bool =  strfind(csv_data.header, 'ACC=T') & strcmp(muse_preset,'ab');
        if ~isempty(acc_bool)
            data.acc.X = X(:,end-2:end) / 1000; %data will be given in 'g' 
            data.acc.channels = channel_labels(end-2:end);
            % Remove ACC rows where all values are Zero
            ix = all(data.acc.X == 0, 2);
            data.acc.X(ix,:) = [];
            data.acc.fs = 50; % Hertz
            data.acc.acquired_samples = size(data.acc.X, 1);
            data.acc.time_vct = (0:data.acc.acquired_samples-1)'/data.acc.fs;
            data.acc.markers = project_markers(data.eeg.time_vct, data.eeg.markers, data.acc.time_vct );
            if ~isempty(csv_data.timestamps)
                data.acc.timestamps = timestamps;
                data.acc.timestamps(ix) = [];
                data.acc.expected_samples = round(data.duration * data.acc.fs);
            end                   
            % All other columns are EEG data
            data.eeg.X = X(:, 1:end-3);
            data.eeg.channels = channel_labels(1:end-3);
        else
            data.eeg.X = X;
            data.eeg.channels = channel_labels;
        end
    end
    
    function data = parse_csv_neuroelectrics_enobio(csv_data)   
    % Parses the CSV_DATA matrix [timestamps, X, marver_vt] according with
    % the HARDWARE = NEUROELECTRICS-ENOBIO
        % Transform X_cell to matrix X, sa each row has the same number of elements
        X = cell2mat(csv_data.X);
        % Get device name
        data.device_name = findfield_inheader(csv_data.header, 'NAME');
        % Get information of the hardware used
        data.hardware = findfield_inheader(csv_data.header, 'HARDWARE');
        % Channel names
        channel_labels = csv_data.channel_names;
        % timestamps and duration
        if ~isempty(csv_data.timestamps)
            timestamps = csv_data.timestamps;
            data.duration = timestamps(end) - timestamps(1);
            channel_labels(1) = [];
        end
        marker_vct = csv_data.marker_vct; 
        channel_labels(end) = [];
    
        data.eeg.X = X;
        data.eeg.fs = 500;
        if ~isempty(csv_data.timestamps)
            data.eeg.timestamps = timestamps;
            data.eeg.expected_samples = round(data.duration * data.eeg.fs);
        end
        data.eeg.acquired_samples = size(data.eeg.X, 1);
        data.eeg.time_vct = (0:data.eeg.acquired_samples-1)'/data.eeg.fs;
        data.eeg.markers = marker_vct;
        data.eeg.channels = channel_labels;
    end

    function data = parse_csv_empatica_winserver(csv_data)   
    % Parses the CSV_DATA matrix [timestamps, X, marver_vt] according with
    % the HARDWARE = EMPATICA-WINSERVER
        % Get device name
        data.device_name = findfield_inheader(csv_data.header, 'NAME');
        % Get information of the hardware used
        data.hardware = findfield_inheader(csv_data.header, 'HARDWARE');
        % timestamps and duration
        if ~isempty(csv_data.timestamps)
            timestamps = csv_data.timestamps;
            data.duration = timestamps(end) - timestamps(1);
        end
        marker_vct = csv_data.marker_vct; 
        % timestamps from Empatica E4
        timestamps_e4 = get_columns_from_cell(csv_data.X, 1);
        data.duration_e4 = timestamps_e4(end) - timestamps_e4(1);
        % stream types 
        stream_types = get_columns_from_cell(csv_data.X, 2);
        
        % Order the data depending of packages
        data_types = {'acc', 'bvp', 'gsr', 'temp', 'ibi', 'bat', 'tag', 'hr'};
        % number of elements per Stream
        datum      = [3,      1,     1,     1,      1,     1,     1,     1];
        % sampling frequency per stream
        fs_types   = [32, 64, 4, 4];
        % Stream, Measurement, Description,     fs
        % 1 E4_Acc     - 3-axis acceleration    @ 32Hz
        % 2 E4_Bvp     - Blood Volume Pulse     @ 64Hz
        % 3 E4_Gsr     - Galvanic Skin Response @ 4Hz
        % 4 E4_Temp    - Skin Temperature       @ 4Hz
        % 5 E4_Ibi     - Interbeat Interval     @ Non-uniform
        % 6 E4_Battery - Device Battery         @ 
        % 7 E4_Tag                              @ Per event (button in device)
        % 8 E4_Hr      - Heart Rate             @
        
        for i_type = 1 : numel(data_types)
            ix = stream_types == i_type;
            tmp_data = vertcat(csv_data.X{ix}); 
            % Timestamps from Empatica
            if ~isempty(tmp_data)
                eval(['data.', data_types{i_type}, '.timestamps_e4', ' = tmp_data(:,1);']);           
                eval(['data.', data_types{i_type}, '.X', ' = tmp_data(:,3:3+datum(i_type)-1);']);            
                eval(['data.', data_types{i_type}, '.markers', ' = project_markers(timestamps_e4, marker_vct, data.', data_types{i_type}, '.timestamps_e4);']);
                if ~isempty(csv_data.timestamps)
                    eval(['data.', data_types{i_type}, '.timestamps', ' = timestamps(ix);']);
                end
            else
                eval(['data.', data_types{i_type}, '.timestamps_e4 = [];']);           
                eval(['data.', data_types{i_type}, '.X = [];']);            
                eval(['data.', data_types{i_type}, '.markers = [];']);
                if ~isempty(csv_data.timestamps)
                    eval(['data.', data_types{i_type}, '.timestamps = [];']);
                end
            end
                
        end
        
        for i_type = 1: numel(fs_types)
            % Sampling frequency
            eval(['data.', data_types{i_type}, '.fs', ' =  fs_types(i_type);']);            
            % Acquired samples
            eval(['data.', data_types{i_type}, '.acquired_samples', ' =  size(data.', data_types{i_type}, '.X ,1);']);
            % Time vector
            eval(['data.', data_types{i_type}, '.time_vct', ' =  (0:data.', data_types{i_type}, '.acquired_samples-1)''/data.', data_types{i_type}, '.fs;']);        
            % Expected samples
            eval(['data.', data_types{i_type}, '.expected_samples', ' = round(data.duration_e4 * data.', data_types{i_type}, '.fs);']);      
        end
        data.bvp.X = -1 * data.bvp.X;
    end
    
    function data = parse_csv_zephyr_bh3raw(csv_data)   
    % Parses the CSV_DATA matrix [timestamps, X, marver_vt] according with
    % the HARDWARE = BH3-RAW
        % Get device name
        data.device_name = findfield_inheader(csv_data.header, 'NAME');
        % Get information of the hardware used
        data.hardware = findfield_inheader(csv_data.header, 'HARDWARE');
        % timestamps and duration
        if ~isempty(csv_data.timestamps)
            timestamps = csv_data.timestamps;
            data.duration = timestamps(end) - timestamps(1);
        end
        marker_vct = csv_data.marker_vct;         
        
        % 32 General info :  3samples   HR,RespRate,SkinTemp  @   1 Hz
        % 33 Respiration  :  18 samples                       @  18 Hz
        % 34 ECG          :  63 samples                       @ 250 Hz
        % 36 RR           :  18 samples                       @  18 Hz
        % 42 ACC          :  60 samples (20 per axes)         @  50 Hz
        
        % General info parsing
        % Resp, ECG, RR and ACC parsing
        data_types = {'resp', 'ecg', 'rr', 'acc'};
        datum       = [18, 63, 18, 60];
        ix_type    = [33, 34, 36, 42 ];
        fs_types   = [18, 250, 18, 50];
        stream_types = get_columns_from_cell(csv_data.X, 1);
        
        for i_type = 1 : numel(data_types)
            ix = stream_types == ix_type(i_type);
            tmp_data = vertcat(csv_data.X{ix});
            tmp_data = tmp_data(:,2:1+datum(i_type));
            n_samples = size(tmp_data,2);           
            tmp_data = reshape(tmp_data',[],1); 
            % Acceleration data is one sample for each axis
            if ix_type(i_type) == 42
                tmp_data = (reshape(tmp_data,3,[]))';
                n_samples = n_samples / 3;
            end
            % Data
            eval(['data.', data_types{i_type}, '.X', ' = tmp_data;']);
            % Acquired samples
            eval(['data.', data_types{i_type}, '.acquired_samples', ' =  size(data.', data_types{i_type}, '.X ,1);']);
            % Sampling frequency
            eval(['data.', data_types{i_type}, '.fs', ' =  fs_types(i_type);']); 
            % Time Vector
            eval(['data.', data_types{i_type}, '.time_vct', ' =  (0:data.', data_types{i_type}, '.acquired_samples-1)''/data.', data_types{i_type}, '.fs;']); 
            
            
            if ~isempty(tmp_data)
                % Timestamps for each stream
                timestamps_tmp = timestamps(ix);
                markers_tmp = project_markers(timestamps, marker_vct, timestamps_tmp);
                % Reshape with datum information
                timestamps_tmp2 = repmat(timestamps_tmp, 1, n_samples );
                timestamps_tmp = reshape(timestamps_tmp2',[],1); 
                markers_tmp2 = repmat(markers_tmp, 1, n_samples );
                markers_tmp2(:,1:end-1) = 0;
                markers_tmp = reshape(markers_tmp2',[],1);
                
                % Timestamps
                eval(['data.', data_types{i_type}, '.timestamps', ' = timestamps_tmp;']);               
                % Markers
                eval(['data.', data_types{i_type}, '.markers', ' = markers_tmp;']);
                % Expected samples
                eval(['data.', data_types{i_type}, '.expected_samples', ' = round(data.duration * data.', data_types{i_type}, '.fs);']);                 
            end        
        end
        
        % ADD PARSE FOR GENERAL PACKAGE
        
        
        
    end
    
    function data = parse_csv_tobii_eyex(csv_data)
    % Parses the CSV_DATA matrix [timestamps, X, marver_vt] according with
    % the HARDWARE = TOBII-EYEX
        % Get device name
        data.device_name = findfield_inheader(csv_data.header, 'NAME');
        % Get information of the hardware used
        data.hardware = findfield_inheader(csv_data.header, 'HARDWARE');
        % Remove timestamps from signals array
        if ~isempty(csv_data.timestamps)
            timestamps = csv_data.timestamps;
            data.duration = timestamps(end) - timestamps(1);
            data.timestamps = timestamps;
        end
        marker_vct = csv_data.marker_vct;         
        timestamps_eyex = get_columns_from_cell(csv_data.X, 1) / 1000;
        data.duration_eyex = timestamps_eyex(end) - timestamps_eyex(1);
        
        % 1 Gaze point   :  X Y                              
        % 3 Eye Position :  XYZ(left) XYZ(right)   
        % 7 Fixation     :  Event(1=start, 3=data, 2=end), X, Y
        
        data_types = {'gaze', 'pos', 'fix'};
        datum       = [2, 6, 3];
        ix_type     = [1, 3, 7];
        stream_types = get_columns_from_cell(csv_data.X, 2);
               
        for i_type = 1 : numel(data_types)         
            ix = stream_types == ix_type(i_type);
            tmp_data = vertcat(csv_data.X{ix});
            tmp_data = [timestamps_eyex(ix), tmp_data(:,3:2+datum(i_type))];         
            % Timestamps from EyeX
            eval(['data.', data_types{i_type}, '.timestamps_eyex', ' = tmp_data(:,1);']);           
            eval(['data.', data_types{i_type}, '.X', ' = tmp_data(:,2:end);']);            
            eval(['data.', data_types{i_type}, '.markers', ' = project_markers(timestamps_eyex, marker_vct, data.', data_types{i_type}, '.timestamps_eyex);']);
            if ~isempty(csv_data.timestamps)
                eval(['data.', data_types{i_type}, '.timestamps', ' = timestamps(ix);']);
            end
            eval(['data.', data_types{i_type}, '.time_vct', ' = data.', data_types{i_type}, '.timestamps_eyex - timestamps_eyex(1);']);
        end      
    end
    
     function data = parse_csv_interaxon_museraw(csv_data)   
    % Parses the CSV_DATA matrix [timestamps, X, marver_vt] according with
    % the HARDWARE = INTERAXON-MUSERAW
        % Get device name
        data.device_name = findfield_inheader(csv_data.header, 'NAME');
        % Get information of the hardware used
        data.hardware = findfield_inheader(csv_data.header, 'HARDWARE');
        % Get MUSE Preset
        muse_preset = findfield_inheader(csv_data.header, 'PRESET');        
        % Remove timestamps from signals array
        if ~isempty(csv_data.timestamps)
            timestamps = csv_data.timestamps;
            data.duration = timestamps(end) - timestamps(1);
        end
        
        marker_vct = csv_data.marker_vct;         
        switch muse_preset
            case {'ab'}
                data.eeg.fs = 500;
            case {'13','14'}
                data.eeg.fs = 220;
        end     
        
        % Order the data depending of packages
        data_types = {'eeg', 'eeg_quantization', 'eeg_dropped_samples', ...
                      'acc', 'acc_dropped_samples', ...
                      'fft0', 'fft1', 'fft2', 'fft3', ...
                      'low_freq_abs', 'delta_abs', 'theta_abs', 'alpha_abs', 'beta_abs', 'gamma_abs', ...
                      'delta_rel', 'theta_rel', 'alpha_rel', 'beta_rel', 'gamma_rel', ...
                      'delta_session', 'theta_session', 'alpha_session', 'beta_session', 'gamma_session', ...
                      'touch_forehead', 'horseshoe', 'is_good', 'blink', 'jaw_clench', ...
                      'concentration', 'mellow', ...
                      'batt', 'drlref'};
        datum      = [4, 1,  1, ...
                      3, 1, ...
                      129, 129, 129, 129, ...
                      4, 4, 4, 4, 4, 4, ...
                      4, 4, 4, 4, 4, ...
                      4, 4, 4, 4, 4, ...
                      1, 4, 4, 1, 1, ...
                      1, 1, ...
                      4, 2 ];
        ix_type   = [0 : 33];
        
        fs_types   = [data.eeg.fs, 0,  0, ...
                      50, 0, ...
                      10, 10, 10, 10, ...
                      10, 10, 10, 10, 10, 10, ...
                      10, 10, 10, 10, 10, ...
                      10, 10, 10, 10, 10, ...
                      10, 10, 10, 10, 10, ...
                      10, 10, ...
                      0.1, 10 ];
        
        stream_types = get_columns_from_cell(csv_data.X, 1);

        
        
        % Availabe data in Muse: http://dev.choosemuse.com/research-tools/available-data
        
        for i_type = 1 : numel(data_types)
            ix = stream_types == ix_type(i_type);
            tmp_data = vertcat(csv_data.X{ix});
            if ~isempty(tmp_data)
                tmp_data = tmp_data(:,2:1+datum(i_type));
                eval(['data.', data_types{i_type}, '.X', ' = tmp_data;']);            
                % Acquired samples
                eval(['data.', data_types{i_type}, '.acquired_samples', ' =  size(data.', data_types{i_type}, '.X ,1);']); 
                if fs_types(i_type) ~= 0
                    % Sampling frequency
                    eval(['data.', data_types{i_type}, '.fs', ' =  fs_types(i_type);']); 
                    % Time Vector
                    eval(['data.', data_types{i_type}, '.time_vct', ' =  (0:data.', data_types{i_type}, '.acquired_samples-1)''/data.', data_types{i_type}, '.fs;']); 
                end          
                if ~isempty(csv_data.timestamps)
                    timestamps_tmp = timestamps(ix);
                    markers_tmp = project_markers(timestamps, marker_vct, timestamps_tmp);
                    eval(['data.', data_types{i_type}, '.timestamps', ' = timestamps(ix);']);
                    eval(['data.', data_types{i_type}, '.markers', ' = markers_tmp;']);
                    if fs_types(i_type) ~= 0
                        % Expected samples
                        eval(['data.', data_types{i_type}, '.expected_samples', ' = round(data.duration * data.', data_types{i_type}, '.fs);']);                 
                    end
                end          
            else
                eval(['data.', data_types{i_type}, '.X = [];']);
                eval(['data.', data_types{i_type}, '.acquired_samples = [];']); 
                eval(['data.', data_types{i_type}, '.fs = [];']);
                eval(['data.', data_types{i_type}, '.time_vct = [];']);
                eval(['data.', data_types{i_type}, '.timestamps = [];']);
                eval(['data.', data_types{i_type}, '.markers = [];']);
                eval(['data.', data_types{i_type}, '.expected_samples = [];']);                 
            end
        end  
    end